# This file is part of Octopus Sensing <https://octopus-sensing.nastaran-saffar.me/>
# Copyright © Nastaran Saffaryazdi 2020
#
# Octopus Sensing is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
#  either version 3 of the License, or (at your option) any later version.
#
# Octopus Sensing is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with Octopus Sensing.
# If not, see <https://www.gnu.org/licenses/>.

from octopus_sensing.devices.shimmer3_streaming import Shimmer3Streaming
from octopus_sensing.devices.brainflow_openbci_streaming import BrainFlowOpenBCIStreaming
from octopus_sensing.devices.camera_streaming import CameraStreaming
from octopus_sensing.devices.audio_streaming import AudioStreaming
from octopus_sensing.devices.open_vibe_streaming import OpenVibeStreaming
from octopus_sensing.devices.brainflow_streaming import BrainFlowStreaming
from octopus_sensing.devices.network_devices.socket_device import SocketNetworkDevice